/*******************************************************************
* MediaCoder - The universal audio/video transcoder
* Copyright (c) 2005-06 Stanley Huang <reachme@citiz.net>
* All rights reserved.
*******************************************************************/

#include "config.h"
#include <stdio.h>
#include "datatypes.h"
#include "transcode.h"
#include "preference.h"

void GetFPSString(int fps, int fpsScale, CString& str)
{
	if (!fps || ! fpsScale) return;
	if (fpsScale == 1) {
		str.AppendFormat("%d", fps);
	} else if (fpsScale & 0x1) {
		str.AppendFormat("%d/%d", fps, fpsScale);
	} else {
		str.AppendFormat("%.3f", (float)fps / fpsScale);
	}
}

void GenAudioFilterOptions(T_ITEM* item, CString& str, int srate)
{
	CPrefBase prefbase;
	CString opts;
	char buf[16];
	int i;
	float f;
	PARAM_OPT afopts[] = {
		{"audiofilter.extra.options", ""},
		{0}
	};

	prefbase.Set("audiofilter.resample");
	if (!srate) {
		srate = pref.GetValueInt(".samplerate");
	}
	if (srate && (!pref.GetValueBool(".downSamplingOnly") || !item->info.audio.samplerate || item->info.audio.samplerate > srate)) {
		if (pref.GetValueBool(".lavcResampler")) {
			opts.AppendFormat(",lavcresample=%d", srate);
		} else {
			opts.AppendFormat(",resample=%d:%d:%d",
				srate,
				pref.GetValueInt(".precious") ? 0 : 1,
				pref.GetValueInt(".method"));
		}
	}
	prefbase.Set("audiofilter.equalizer");
	if (pref.GetValueBool(".enabled")) {
		opts += ",equalizer=";
		for (i = 0; i <= 9; i++) {
			sprintf(buf, ".band%d", i);
			opts.AppendFormat("%d:", pref.GetValueInt(buf));
		}
	}
	prefbase.Set("audiofilter.channels");
	if (pref.GetValueBool(".enabled")) {
		int nr=0;
		for (i = 0; i < 5; i++) {
			sprintf(buf, ".channel%d", i);
			if (pref.GetValueInt(buf) > 0) nr++;
		}
		opts.AppendFormat(",channels=%d:%d:", pref.GetValueInt(".channelNum"), nr);
		for (i=0;i<5;i++) {
			int ch;
			sprintf(buf, ".channel%d", i);
			if ((ch = pref.GetValueInt(buf))>0) {
				opts.AppendFormat("%d:%d:", ch-1, i);
			}
		}
	}
	f = pref.GetValueFloat("audiofilter.surround.delay");
	if (f) {
		opts.AppendFormat(",surround=%.2f", f);
	}
	prefbase.Set("audiofilter.delay");
	if (pref.GetValueBool(".enabled")) {
		opts += ",delay=";
		for (i = 0; i <= 6; i++) {
			sprintf(buf, ".channel%d", i);
			opts.AppendFormat("%.2f:", pref.GetValueFloat(buf));
		}
	}
	prefbase.Set("audiofilter.extraStereo");
	if (pref.GetValueBool(".enabled")) {
		opts.AppendFormat(",extrastereo=%.2f", pref.GetValueFloat(".coeff"));
	}

	prefbase.Set("audiofilter.volume");
	f = pref.GetValueFloat(".gain");
	// workaround for AMR sound level too low issue
	if (item->target.audioformat == AC_AMR) {
		f += 3;
	}
	if (pref.GetValueBool(".normalization")) {
		opts += ",volnorm=2";
	} else if (f) {
		opts.AppendFormat(",volume=%.2f:%d",
			f,
			pref.GetValueInt(".clipping"));
	}

	if (pref.GetValueBool("audiofilter.compressor.enabled")) {
		opts += ",comp";
	}
	pref.GenOptions(opts, afopts, ',');
	if (!opts.IsEmpty()) {
		str += " -af ";
		str += opts.Mid(1);
	}
}

void GenVideoFilterOptions(T_ITEM* item, CString& str)
{
	CPrefBase prefbase;
	CString opts;
	PARAM_OPT vfopts[] = {
		{"videofilter.screenshot.enabled", "screenshot"},
		{"videofilter.extra.options", ""},
		{0}
	};

	int i;
	prefbase.Set("videofilter.crop");
	if (pref.GetValueBool(".enabled")) {
		int x = item->target.crop.x;
		int y = item->target.crop.y;
		int w = item->target.crop.w;
		int h = item->target.crop.h;
		if (!(x | y | w | h)) {
			x = pref.GetValueInt(".left");
			y = pref.GetValueInt(".top");
			w = pref.GetValueInt(".width");
			h = pref.GetValueInt(".height");
		}
		if (x | y | w | h) {
			int ex = 0, ey = 0, ew = 0, eh = 0;
			if (x < 0) {
				ex = -x;
				ew = w;
				w += x;
				x = 0;
			}
			if (y < 0) {
				ey = -y;
				eh = h;
				h += y;
				y = 0;
			}
			if (w > item->info.video.width) {
				if (!ew) ew = w;
				w = item->info.video.width;
			}
			if (h > item->info.video.height) {
				if (!eh) eh = h;
				h = item->info.video.height;
			}
			opts += ",crop=";
			if (w) opts.AppendFormat("%d", w);
			opts += ':';
			if (h) opts.AppendFormat("%d", h);
			opts += ':';
			if (x) opts.AppendFormat("%d", x);
			opts += ':';
			if (y) opts.AppendFormat("%d", y);
			if (ex | ey | ew | eh) {
				opts += ",expand=";
				if (ew) opts.AppendFormat("%d", ew);
				opts += ':';
				if (eh) opts.AppendFormat("%d", eh);
				opts += ':';
				if (ex) opts.AppendFormat("%d", ex);
				opts += ':';
				if (ey) opts.AppendFormat("%d", ey);
			}
		}
	}

	switch (pref.GetValueInt("videofilter.itf.type")) {
	case 1:
		opts += ",detc";
		break;
	case 2:
		opts += ",pullup";
		break;
	}

	prefbase.Set("videofilter.scale");
	if (pref.GetValueBool(".enabled")) {
		static const char* alg[] = {":", "0.00:0.75", "0.00:0.50", "0.33:0.33", "1.00:0.00"};
		int w = pref.GetValueInt(".width");
		int h = pref.GetValueInt(".height");
		if (w & h) {
			opts += ",scale=";
			opts.AppendFormat("%d:%d:%d:%d:%s", 
				w, h,
				pref.GetValueInt(".interlaced"),
				pref.GetValueInt(".chroma"),
				alg[pref.GetValueInt(".algorithm")]);
		}

	}
	prefbase.Set("videofilter.expand");
	if (pref.GetValueBool(".enabled")) {
		static const char* aspects[] = {"3/4", "9/16", "1:2"};
		opts += ",expand=";
		if (i = pref.GetValueInt(".width")) opts.AppendFormat("%d", i);
		opts += ':';
		if (i = pref.GetValueInt(".height")) opts.AppendFormat("%d", i);
		opts += ':';
		if (i = pref.GetValueInt(".x")) opts.AppendFormat("%d", i);
		opts += ':';
		if (pref.GetValueInt(".y")) opts.AppendFormat("%d", i);
		opts.AppendFormat("%d", pref.GetValueBool(".osd"));
		opts += ':';
		if (i = pref.GetValueInt(".aspect")) {
			opts += aspects[i - 1] + 1;
		}
		opts += ':';
		if ((i = pref.GetValueInt(".roundup")) > 1) {
			opts.AppendFormat("%d", i);
		}
		opts += ',';
	}
	prefbase.Set("videofilter.postproc");
	if (i = pref.GetValueInt(".mode")) {
		char* ppmode[]={":c/", ":n/", ":y/"};
		CString ppopts;
		PARAM_OPT subopts[] = {
			{".level", "|al|al:f"},
			{".hdeblock", "ha"},
			{".vdeblock", "va"},
			{".dering", "dr"},
			{".denoiser", "tn:1:2:3"},
			{".deint", "|lb|li|ci|md|fd|l5"},
			{0}
		};
		pref.GenOptions(ppopts, subopts, ppmode[i - 1]);
		if (!ppopts.IsEmpty()) {
			opts += ",pp=";
			opts += ppopts.Mid(3);
		}
	}
	prefbase.Set("videofilter.denoise");
	if ((i = pref.GetValueInt("videofilter.denoise.mode"))) {
		opts.AppendFormat(",%s=%d:%d:%d",
			i == 1 ? "denoise3d" : "hqdn3d",
			pref.GetValueInt(".luma"),
			pref.GetValueInt(".chroma"),
			pref.GetValueInt(".strength"));
	}
	if (pref.GetValueBool("overall.video.harddup") || (item->target.container != CF_DEFAULT && item->target.container != CF_AVI && item->target.container != CF_MKV)) {
		opts += ",harddup";
	}
	if (pref.GetValueInt("videofilter.rotate.enabled")) {
		opts.AppendFormat(",rotate=%d", pref.GetValueInt("videofilter.rotate.mode"));
	}
	float f;
	CString subopts;
	if ((f = pref.GetValueFloat("videofilter.eq.gamma")) != 1) subopts.Format("%.4f", f);
	subopts += ':';
	if ((i = pref.GetValueInt("videofilter.eq.contrast")) != 50) subopts.AppendFormat("%.4f", (float)i / 50);
	subopts += ':';
	if ((i = pref.GetValueInt("videofilter.eq.brightness")) != 0) subopts.AppendFormat("%.4f", (float)i / 100);
	subopts += ':';
	if ((f = pref.GetValueFloat("videofilter.eq.saturation")) != 1) subopts.AppendFormat("%.4f", f);
	if (subopts != ":::") {
		opts += ",eq2=";
		opts += subopts;
	}
	if ((i = pref.GetValueInt("videofilter.eq.hue")) != 0) opts.AppendFormat(",hue=%d", i);
	pref.GenOptions(opts, vfopts, ',');
	if (!opts.IsEmpty()) {
		str += " -vf " + opts.Mid(1, opts.GetLength() - 1);
	}
}

void GenCommonOptions(T_ITEM* item, CString& str, int enc)
{
	PARAM_OPT opts[] = {
		{"overall.decoding.slang", "-slang "},
		{"overall.decoding.alang", "-slang "},
		{"overall.decoding.aid", "-aid "},
		{"overall.decoding.vid", "-vid "},
		{"overall.decoding.channels", "-channels "},
		{"overall.decoding.audioFile", "-audiofile "},
		{"overall.decoding.rebuildAVIIndex", "-forceidx"},
		{0}
	};
	
	if (!enc) {
		PARAM_OPT mpopts[] = {
			{"overall.mplayer.cache", "-cache "},
			{"overall.mplayer.options", ""},
			{0}
		};
		str.AppendFormat(" -monitoraspect %d:%d", mcinst.scrw, mcinst.scrh);
		pref.GenOptions(str, mpopts, ' ');
	} else {
		PARAM_OPT meopts[] = {
			{"container.avi.noodml", "-noodml"},
			{"overall.video.noSkip", "-noskip"},
			{0}
		};
		if (mcinst.duration > 0) str.AppendFormat(" -endpos %d", mcinst.duration / 1000);
		pref.GenOptions(str, meopts, ' ');
	}
	if (mcinst.startpos) str.AppendFormat(" -ss %d", mcinst.startpos / 1000);

	if (item && item->info.discroot) {
		str.AppendFormat(" -dvd-device \"%s\"", item->info.discroot);
	}
	if (item && item->info.chapter) {
		str.AppendFormat(" -chapter %d", item->info.chapter);
	}
	if (item->subfile && *item->subfile) {
		str.AppendFormat(" -sub \"%s\"", item->subfile);
	}
	pref.GenOptions(str, opts, ' ');
}
