/*******************************************************************
* MediaCoder - The universal audio/video transcoder
* Copyright (c) 2005-06 Stanley Huang <reachme@citiz.net>
* All rights reserved.
*******************************************************************/

#include "config.h"
#include <stdio.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <iconv.h>
#include "atlstr.h"
#include "preference.h"
#include "resource.h"
#include "paramlistdef.h"
#include "transcode.h"
#include "xmlaccessor.h"

CPrefBase::CPrefBase(char* path)
{
	prevnode = pref.basenode;
	if (path) pref.SetBase(path);
}

void CPrefBase::Set(char* path)
{
	pref.SetBase(path);
}

CPrefBase::~CPrefBase()
{
	pref.basenode = prevnode;
}

int CPrefHelper::GetValueInt(char* key)
{
	PARAM_DATA* pd = GetEntry(key);
	return pd ? (int)pd->value : 0;
}

char* CPrefHelper::GetValueStr(char* key)
{
	PARAM_DATA* pd = GetEntry(key);
	return pd ? (char*)pd->value : "";
}

bool CPrefHelper::GetValueBool(char* key)
{
	PARAM_DATA* pd = GetEntry(key);
	return pd ? (pd->value ? true : false) : false;
}

float CPrefHelper::GetValueFloat(char* key)
{
	PARAM_DATA* pd = GetEntry(key);
	return pd ? *(float*)pd->value : 0;
}

char* CPrefHelper::GetValueEnum(PARAM_DATA* pd, int index)
{
	if (index < 0) index = (int)pd->value;
	char *p=(char*)pd->enums;
	int i=0;
	do {
		if (i++ == index) {
			if (*p == '\t' || *p == ' ') p++;
			return p;
		}
		while (*(p++));
	} while (*p);
	return 0;
}

char* CPrefHelper::GetValueEnum(char* key, int index)
{
	PARAM_DATA* pd = GetEntry(key);
	if (!pd) return "";
	return GetValueEnum(pd, index);
}

void CPrefHelper::SetValue(PARAM_DATA* pd, void* value)
{
	switch (pd->type) {
	case PT_INT:
	case PT_BOOL:
	case PT_ENUM:
		SetValue(pd, (int)value);
		break;
	case PT_STRING:
	case PT_FILE:
	case PT_DIR:
		SetValue(pd, (char*)value);
		break;
	case PT_FLOAT:
		SetValue(pd, value ? *(float*)value : 0);
		break;
	}
}

void CPrefHelper::SetValue(char* key, void* value)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) SetValue(pd, value);
}

void CPrefHelper::SetValue(char* key, int value)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) SetValue(pd, value);
}

void CPrefHelper::SetValue(char* key, char* value)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) SetValue(pd, value);
}

void CPrefHelper::SetValue(char* key, CString& value)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) SetValue(pd, value.GetBuffer());
}

void CPrefHelper::SetValue(char* key, bool value)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) pd->value = value ? (void*)1 : 0;
}

void CPrefHelper::SetValue(char* key, float value)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) SetValue(pd, value);
}

bool CPrefHelper::IsDefaultValue(PARAM_DATA* pd)
{
	switch (pd->type) {
	case PT_FILE:
	case PT_DIR:
	case PT_STRING:
		if ((!pd->valueDefault && !(char*)pd->value) || (pd->valueDefault && !strcmp((char*)pd->value,(char*)pd->valueDefault))) return true;
		break;
	case PT_FLOAT:
		if ((pd->valueDefault && *(float*)pd->value == *(float*)pd->valueDefault) ||
			!pd->valueDefault && *(float*)pd->value == 0) return true;
		break;
	case PT_INT:
	case PT_ENUM:
	case PT_BOOL:
		if (pd->value == pd->valueDefault) return true;
		break;
	default:
		return true;
	}
	return false;
}

bool CPrefHelper::IsDefaultValue(char* key)
{
	PARAM_DATA* pd = GetEntry(key);
	return pd ? IsDefaultValue(pd) : false;
}

void CPrefHelper::SetDefaultValue(char* key, void* value, bool setval)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) SetDefaultValue(pd, value, setval);
}

void CPrefHelper::SetDefaultValue(PARAM_DATA* pd, void* value, bool setval)
{
	switch (pd->type) {
	case PT_FILE:
	case PT_DIR:
	case PT_STRING:
		if (pd->valueDefault) free(pd->valueDefault);
		pd->valueDefault = (void*)strdup((char*)value);
		break;
	case PT_FLOAT:
		if (!pd->valueDefault) pd->valueDefault = malloc(sizeof(float));
		*(float*)pd->valueDefault = *(float*)value;
		break;
	default:
		pd->valueDefault = value;
	}
	if (setval) SetValue(pd, value);

}

PARAM_DATA* CPrefHelper::SetBase(char* path)
{
	if (!path) {
		basenode = 0;
		return 0;
	}

	int i = 0;
	int len;
	PARAM_DATA* pd = root.sub;
	char *s = path;
	char *p;
	basenode = 0;

	for (; pd; s = p + 1) {
		for (p = s; *p && *p != '.' && *p !='/'; p++);
		len = p - s;
		for (i = 0; pd[i].type && strncmp(s, pd[i].key, len); i++);
		if (!pd[i].type) {
			// no matched node found
#ifdef _DEBUG
			MessageBox(0, path, "Invalid path", MB_OK);			
#endif
			return 0;
		}
		pd = pd[i].sub;
		if (!*p) {
			return (basenode = pd);
		}
	}
	return 0;
}

void CPrefHelper::SetEnums(char* key, char* enums)
{
	PARAM_DATA* pd = GetEntry(key);
	if (pd) {
		pd->enums = enums;
	}
}

PARAM_DATA* CPrefHelper::GetEntry(char* key)
{
	if (!key) return &root;
	int i = 0;
	int len;
	PARAM_DATA* pd = root.sub;
	char *s = key;
	char *p;
	if (key[0] == '.' && basenode) {
		s++;
		pd = basenode;
	}
	for (; pd; s = p + 1) {
		for (p = s; *p && *p != '.' && *p !='/'; p++);
		len = p - s;
		for (i = 0; pd[i].type && !(!pd[i].key[len] && !memcmp(s, pd[i].key, len)); i++);
		if (!pd[i].type) {
			// no matched node found
#ifdef _DEBUG
			MessageBox(0, key, "Invalid key", MB_OK);			
			return 0;
#endif
			return 0;
		}
		if (!*p) {
			return &pd[i];
		}
		pd = pd[i].sub;
	}
#ifdef _DEBUG
	MessageBox(0, key, "Invalid key", MB_OK);			
#endif
	return 0;
}

void CPrefHelper::GenOptions(CString& cmd, PARAM_OPT paramopts[])
{
	GenOptions(cmd, paramopts, " ");
}

void CPrefHelper::GenOptions(CString& cmd, PARAM_OPT paramopts[], int delimiter)
{
	GenOptions(cmd, paramopts, (char*)&delimiter);
}

void CPrefHelper::GenOptions(CString& cmd, PARAM_OPT paramopts[], char* delimiter)
{
	for (int i = 0; paramopts[i].key; i++) {
		PARAM_DATA* param = GetEntry(paramopts[i].key);
		if (!param) continue;
		char* opt = paramopts[i].opt;
		switch (param->type) {
		case PT_BOOL: {
			char *p = strchr(opt, '|');
			if (p) {
				cmd += delimiter;
				if (param->value) {
					cmd += (p + 1);
				} else {
					for (p = opt; *p != '|'; p++) cmd += *p;
				}
			} else if (param->value) {
				cmd += delimiter;
				cmd += opt;
			}
			} break;
		case PT_ENUM:
		case PT_INT: {
			char *p = strchr(opt, '|');
			if (p) {
				p = opt;
				for (int n = 0; n < (int)param->value; n++, p++) {
					if (!(p = strchr(p, '|'))) break;
				}
				if (p && *p && *p != '|') {
					cmd += delimiter;
					for (; *p && *p != '|'; p++) cmd += *p;
				}
			} else if (param->value || param->value != param->valueDefault) {
				cmd += delimiter;
				cmd.AppendFormat("%s%d", opt, param->value);
			}
			} break;
		case PT_STRING:
		case PT_FILE:
		case PT_DIR:
			if (*(char*)param->value) {
				cmd += delimiter;
				cmd.AppendFormat("%s%s", opt, param->value);
			}
			break;
		case PT_FLOAT: {
			cmd += delimiter;
			cmd.AppendFormat("%s%f", opt, *(float*)param->value);
			} break;
		}
	}
}

int CPrefHelper::GetStringValue(PARAM_DATA* pd, char* buf, int bufsize)
{
	int ret = 0;
	switch (pd->type) {
	case PT_ENUM:
		return _snprintf(buf, bufsize, "%s", GetValueEnum(pd));
	case PT_INT:
		return _snprintf(buf, bufsize, "%d", pd->value);
	case PT_BOOL:
		return _snprintf(buf, bufsize, "%s", pd->value ? "true" : "false");
	case PT_DIR:
	case PT_FILE:
	case PT_STRING:
		return _snprintf(buf, bufsize, "%s", pd->value);
	case PT_FLOAT:
		return _snprintf(buf, bufsize, "%f", *(float*)pd->value);
	default:
		buf[0] = 0;
		return 0;
	}
}

int CPrefHelper::GetStringValue(char* key, char* buf, int bufsize)
{
	PARAM_DATA *pd = GetEntry(key);
	if (!pd) {
		buf[0] = 0;
		return 0;
	}
	return GetStringValue(pd, buf, bufsize);
}

void CPrefHelper::GetStringValue(PARAM_DATA* pd, CString& str)
{
	switch (pd->type) {
	case PT_ENUM:
		str = GetValueEnum(pd);
		break;
	case PT_INT:
		str.Format("%d", pd->value);
		break;
	case PT_BOOL:
		str = pd->value ? "true" : "false";
		break;
	case PT_DIR:
	case PT_FILE:
	case PT_STRING:
		str = (char*)pd->value;
		break;
	case PT_FLOAT:
		str.Format("%f", *(float*)pd->value);
		break;
	default:
		str = "";
	}
}

void CPrefHelper::GetStringValue(char* key, CString& str)
{
	PARAM_DATA *pd = GetEntry(key);
	if (!pd) {
		str = "";
		return;
	}
	GetStringValue(pd, str);
}

void CPrefHelper::SetStringValue(PARAM_DATA* pd, char* value)
{
	switch (pd->type) {
	case PT_ENUM:
		if (!value || !*value) {
			pd->value = 0;
			break;
		}
		for (int i = 0; ;i++) {
			char *s = GetValueEnum(pd, i);
			if (!s) break;
			if (!stricmp(value, s)) {
				pd->value = (void*)i;
				break;
			}
		}
		break;
	case PT_INT:
		pd->value = value ? (void*)atoi(value) : 0;
		break;
	case PT_BOOL:
		pd->value = value ? ((!stricmp(value, "true") || !strcmp(value, "1")) ? (void*)1 : 0) : 0;
		break;
	case PT_DIR:
	case PT_FILE:
	case PT_STRING:
		if (value) {
			char *p = (char*)pd->value;
			int len = strlen(value);
			if (len > pd->max) len = pd->max;
			memcpy(p, value, len);
			if (pd->type == PT_DIR && p[len - 1] != '\\') {
				 p[len++] = '\\';
				 if (len > pd->max) len = pd->max;
			}
			p[len] = 0;
		} else {
			*(char*)pd->value = 0;
		}
		break;
	case PT_FLOAT:
		*(float*)pd->value = value ? (float)atof(value) : 0;
		break;
	}
}

void CPrefHelper::SetStringValue(char* key, char* value)
{
	PARAM_DATA *pd = GetEntry(key);
	if (pd) SetStringValue(pd, value);
}

void CPrefHelper::Revert(PARAM_DATA* pd)
{
	SetValue(pd, pd->valueDefault);
}

void CPrefHelper::Revert(char* key)
{
	PARAM_DATA *pd = GetEntry(key);
	if (pd) Revert(pd);
}

int CPrefHelper::Init(char* xmlfile)
{
	if (xml.Open(xmlfile) || !xml.isMatched(rootNodeName)) return -1;
	memset(&root, 0, sizeof(root));
	root.key = rootNodeName;
	root.type = PT_ROOT;
	InitParams(&root);
	return 0;
}

static char* typeNames[] = {"", "int", "string", "enum", "bool", "float", "file", "dir", "node", "catalog", "root", 0};

void CPrefHelper::InitParams(PARAM_DATA* pd)
{
	int count = xml.getChildCount();
	if (count == 0) return;
	xmlNodePtr node = xml.goChild();
	if (xml.isMatched("desc")) {
		pd->desc = strdup(xml.getAttribute("text"));
		node = xml.goNext();
		count--;
	}
	if (!node) return;
	pd->sub = (PARAM_DATA*)calloc(count + 1, sizeof(PARAM_DATA));
	pd = pd->sub;
	for (int n = 0; n < count && node; n++, node = xml.goNext()) {
		if (!xml.isMatched("node")) continue;
		for (xmlAttrPtr attr = node->properties; attr; attr = attr->next) {
			if (!attr->children) continue;
			char* name = (char*)attr->name;
			char* value = (char*)attr->children->content;
			if (!strcmp(name, "type")) {
				for (int i = 1; typeNames[i]; i++) {
					if (!strcmp(value, typeNames[i])) {
						pd[n].type = i;
						break;
					}
				}
			} else if (!strcmp(name, "key")) {
				pd[n].key = strdup(value);
			} else if (!strcmp(name, "text")) {
				pd[n].text = strdup(value);
			} else if (!strcmp(name, "min")) {
				pd[n].min = atoi(value);
			} else if (!strcmp(name, "max")) {
				pd[n].max = atoi(value);
			}
		}
		switch (pd[n].type) {
		case PT_CATALOG:
		case PT_NODE:
			InitParams(&pd[n]);
			continue;
		case PT_STRING:
		case PT_FILE:
		case PT_DIR:
			pd[n].value = malloc(pd[n].max + 1);
			*(char*)pd[n].value = 0;
			break;
		case PT_FLOAT:
			pd[n].value = malloc(sizeof(float));
			*(float*)pd[n].value = 0;
			break;
		}
		if (xml.goChild()) {
			int len = 0;
			char* s = 0;
			char* v = 0;
			int enumidx = 0;
			do {
				char* name = xml.getNodeName();
				char* value = xml.getNodeValue();
				if (!strcmp(name, "enum")) {
					int vlen = strlen(value) + 1;
					s = (char*)realloc(s, len + vlen + 1);
					for (int i = 0; i < vlen; i++) {
						s[len + i] = value[i] == '|' ? 0 : value[i];
					}
					len += vlen;
					s[len] = 0;
					if (xml.getAttribute("sel")) {
						SetDefaultValue(&pd[n], (void*)enumidx, true);
					}
					enumidx++;
				} else if (!strcmp(name, "value")) {
					v = value;
				} else if (!strcmp(name, "desc")) {
					pd[n].desc = strdup(xml.getAttribute("text"));
				} else if (!strcmp(name, "flag")) {
					if (xml.getAttribute("nosave")) {
						pd[n].flags |= PF_NOSAVE;
					} else if (xml.getAttribute("norevert")) {
						pd[n].flags |= PF_NOREVERT;
					}
				}
			} while (xml.goNext());
			if (s) {
				s[len] = 0;
				pd[n].enums = s;
			}
			if (v) {
				SetStringValue(&pd[n], v);
				SetDefaultValue(&pd[n], pd[n].value);
			}
			xml.goParent();
		}
	}
	xml.goParent();
}

void CPrefHelper::Destroy()
{
	DestroyParams(root.sub);
	free(root.sub);
}

void CPrefHelper::DestroyParams(PARAM_DATA* params)
{
	for (; params->type; params++) {
		if (params->sub)
			DestroyParams(params->sub);
		switch (params->type) {
		case PT_FILE:
		case PT_DIR:
		case PT_STRING:
		case PT_FLOAT:
			if (params->value) free(params->value);
			if (params->valueDefault) free(params->valueDefault);
			break;
		}
		if (params->enums) free(params->enums);
		if (params->text) free(params->text);
		if (params->key) free(params->key);
		if (params->desc) free(params->desc);
	}
}

void CPrefHelper::Secharset(const char* tocode, const char* fromcode)
{
	if (!tocode && !fromcode) {
		if (cd) {
			iconv_close(cd);
			cd = 0;
		}
	} else {
		cd = iconv_open(tocode ? tocode : "", fromcode ? fromcode : "");
	}
}

char* CPrefHelper::Convercharset(char* str, int len)
{
	size_t insize = (len ? len : strlen(str)) + 1;
	size_t outsize = insize * 2;
	char* buf = (char*)malloc(outsize);
	char* out = buf;
	char* in = str;
	iconv(cd, (const char**)&in, &insize, &out, &outsize);
	return buf;
}

void CPrefHelper::SaveParams(PARAM_DATA* pd, int flags, int level)
{
	if (!pd) return;
	if (pd->type != PT_ROOT) {
		xml.addChild("node");
		xml.addAttribute("key", pd->key);
		if (flags & FLAG_TYPE) xml.addAttribute("type", typeNames[pd->type]);
		if (flags & FLAG_TEXT) xml.addAttribute("text", pd->text);
	}
	if (pd->desc && (flags & FLAG_DESC)) {
		xml.addChild("desc", pd->desc);
		xml.goParent();
	}
	if (pd->sub && (unsigned int)level > 1) {
		pd = pd->sub;
		for (; pd->type; pd++) {
			SaveParams(pd, flags, level - 1);
		}
	} else if (flags & FLAG_SAVE_ALL || (!pref.IsDefaultValue(pd) && !(pd->flags & PF_NOSAVE))){
		if (flags & FLAG_RANGE) {
			char buf[16];
			if (pd->min) {
				sprintf(buf, "%d", pd->min);
				xml.addAttribute("min", buf);
			}
			if (pd->max) {
				sprintf(buf, "%d", pd->max);
				xml.addAttribute("max", buf);
			}
		}
		if (pd->enums && (flags & FLAG_ENUM)) {
			for (int i = 0; ;i++) {
				char *p = GetValueEnum(pd, i);
				if (!p) break;
				xml.addChild("enum", p);
				xml.goParent();
			}
		}
		if (flags & FLAG_VALUE) {
			CString str;
			GetStringValue(pd, str);
			char* buf = Convercharset(str.GetBuffer(), str.GetLength());
			if (*buf) {
				xml.addAttribute("value", buf);
			}
			free(buf);
		}
	}
	xml.goParent();
}

int CPrefHelper::Save(char* filename)
{
	xml.New("1.0", rootNodeName);
	Secharset("UTF-8");
	SaveParams(&root, FLAG_VALUE);
	xml.Save(filename, "UTF-8");
	Secharset();
	char *buf;
	int bytes = xml.Dump(&buf, "UTF-8");
	return 0;
}

void CPrefHelper::LoadValue(char* file, char* key)
{
	Load(file, key);
}

void CPrefHelper::AdjustValue(PARAM_DATA* pd, char* value)
{
	switch (value[0]) {
	case '+':
		switch (pd->type) {
		case PT_ENUM:
		case PT_INT:
			*(int*)pd->value += + atoi(value + 1);
			break;
		case PT_FLOAT:
			*(float*)pd->value += (float)atof(value + 1);
			break;
		case PT_STRING:
			strncpy((char*)pd->value, value + 1, pd->max - strlen((char*)pd->value));
			break;
		}
		break;
	case '-':
		switch (pd->type) {
		case PT_ENUM:
		case PT_INT:
			*(int*)&pd->value -= atoi(value + 1);
			break;
		case PT_FLOAT:
			*(float*)pd->value -= (float)atof(value + 1);
			break;
		}
		break;
	case '*':
		switch (pd->type) {
		case PT_INT:
			pd->value = (void*)(int)(atof(value + 1) * (int)pd->value);
			break;
		case PT_FLOAT:
			*(float*)pd->value *= (float)atof(value + 1);
			break;
		}
		break;
	}
}

void CPrefHelper::LoadParams(char* path)
{
	PARAM_DATA *pd = GetEntry(path);
	char *p;
	if (!pd) return;
	p = xml.getAttribute("value");
	if (p) {
		if (*p == '~')
			AdjustValue(pd, p + 1);
		else
			SetStringValue(pd, p);
	}
	if (!pd->sub) {
		if (!xml.goChild()) {
			if (!p) SetValue(pd, pd->valueDefault);
			return;
		}
		for (;;){
			if (xml.isMatched("value")) {
				p = xml.getNodeValue();
				if (p) {
					p = Convercharset(p);
					SetStringValue(pd, p);
					free(p);
				}
			}
			if (!xml.goNext()) {
				if (!p) SetValue(pd, pd->valueDefault);
				break;
			}
		}
		xml.goParent();
	}
	if (xml.goChild()) {
		CString key;
		do {
			if (!xml.isMatched("node")) continue;
			key = "";
			if (path) {
				key = path;
				key += '.';
			}
			key += xml.getAttribute("key");
			LoadParams(key.GetBuffer());
		} while (xml.goNext());
		xml.goParent();
	}
}

int CPrefHelper::Load(char* filename, char* key, char* rootpath)
{
	if (xml.Open(filename)) return -1;
	if (rootpath) {
		if (!xml.goToKey(rootpath)) return -1;
	} else  {
		if (rootNodeName) {
			if (!xml.isMatched(rootNodeName)) return -1;
		}
		if (key) {
			char *s = strdup(key);
			char *p = s;
			char *dot;
			xml.goChild();
			for (;;) {
				dot = strchr(p, '.');
				if (dot) *dot = 0;
				if (!xml.findNode("node", "key", p) || !dot || !xml.goChild()) break;
				p = dot + 1;
			}
			free(s);
		}
	}
	Secharset(0,"UTF-8");
	LoadParams(key);
	Secharset();
	return 0;
}

int CPrefHelper::ReadXml(char* buf, char* baskey)
{
	if (!buf) return -1;
	if (xml.Read(buf, "pref.xml", "UTF-8")) return -1;
	if (baskey) {
		if (!xml.goToKey(baskey)) return -1;
	}
	Secharset(0,"UTF-8");
	LoadParams(0);
	Secharset();
	return 0;
}

int CPrefHelper::WriteXml(char** pbuf, char* basekey, int level, int flags)
{
	xml.New("1.0", rootNodeName);
	Secharset("UTF-8");
	if (basekey) {
		PARAM_DATA* pd;
		char *s = strdup(basekey);
		char *p = s;
		char *dot;
		for (;;) {
			dot = strchr(p, '.');
			if (!dot) break;
			*dot = 0;
			pd = GetEntry(s);
			if (!pd) break;
			xml.addChild("node");
			xml.addAttribute("type", typeNames[pd->type]);
			xml.addAttribute("text", pd->text);
			xml.addAttribute("key", pd->key);
			*dot = '.';
			p = dot + 1;
		}
		free(s);
		pd = GetEntry(basekey);
		if (pd) SaveParams(pd, flags, level);
	} else {
		SaveParams(&root, flags, level);
	}
	Secharset();
	int bytes = xml.Dump(pbuf, "UTF-8");
	return bytes;
}

// exported APIs

int mcGetPrefProperty(char* key, PARAM_PROPERTY property, void* pval)
{
	PARAM_DATA* entry = pref.GetEntry(key);
	if (!entry) return -1;
	if (entry->type == 0) return -1;
	switch (property) {
	case PP_KEY: *(char**)pval = entry->key; break;
	case PP_TYPE: *(int*)pval = entry->type; break;
	case PP_RANGE:
		*(int*)pval = entry->min;
        *((int*)pval + 1) = entry->max;
		break;
	case PP_CAPTION: *(char**)pval = entry->text; break;
	case PP_DESC:
		break;
	default:
		return -1;
	}
	return 0;
}

void* mcGetPrefValue(char* key)
{
	PARAM_DATA* pd = pref.GetEntry(key);
	return pd ? pd->value : 0;
}

int mcGetPrefValueInt(char* key)
{
	return pref.GetValueInt(key);
}

char* mcGetPrefValueStr(char* key)
{
	return pref.GetValueStr(key);
}

int mcGetPrefValueStrMax(char* key)
{
	PARAM_DATA* pd = pref.GetEntry(key);
	return pd ? pd->max : 0;
}

char* mcGetPrefValueEnum(char* key, int index)
{
	return pref.GetValueEnum(key, index);
}

float mcGetPrefValueFloat(char* key)
{
	return pref.GetValueFloat(key);
}

void mcSetPrefValue(char* key, void* value)
{
	pref.SetValue(key, value);
}

void mcSetPrefValueFloat(char* key, float value)
{
	pref.SetValue(key, value);
}

void mcSetPrefValueInt(char* key, int value)
{
	pref.SetValue(key, value);
}

void mcSetPrefValueStr(char* key, char* value)
{
	pref.SetValue(key, value);
}

int mcSetPrefValueXML(char* xml, char* basekey)
{
	if (!xml) return -1;
	if ((xml[1] == ':' && xml[2] == '\\') || xml[0] == '/' || xml[1] == '/') {
		return pref.Load(xml, 0, basekey);
	} else {
		return pref.ReadXml(xml, basekey);
	}
}

char* mcGetPrefValueXML(char* key, int level)
{
	char *xmltext;
	return pref.WriteXml(&xmltext, key, level) > 0 ? xmltext : 0;
}
